/**
 * await.h
 */
#ifndef __AWAIT_INCLUDED
#define __AWAIT_INCLUDED

#include <await/export.hdr>
#include <await/lockmanager.h>

namespace await_n {

enum { FOREVER = -1 };

class SynchObject;
class AWAIT_API SynchHandle {
private:
	SynchObject *m_pObject;
public:
	SynchHandle();
	~SynchHandle();
	bool cont();
	void next();
};

class AwaitSwitchObject;
class AWAIT_API AwaitSwitchHandle {
private:
	AwaitSwitchObject *m_pObject;
public:
	AwaitSwitchHandle(long);
	~AwaitSwitchHandle();
	bool cont();
	void next();
	bool check(bool);
	bool may();
	bool timedout();
};

class LockSwitchObject;
class AWAIT_API LockSwitchHandle {
private:
	LockSwitchObject *m_pObject;
public:
	LockSwitchHandle(long, ExtraInfo const &);
	~LockSwitchHandle();
	bool cont();
	void next();
	bool prep();
	void collect(LockSlot const &);
	bool selected();
	bool timedout();
	bool victimed();
};

} // namespace await_n

#define synch() FOR(await_n::SynchHandle theAuxiliaryObject; theAuxiliaryObject.cont(); theAuxiliaryObject.next())


#define await_switch(timeout) FOR(await_n::AwaitSwitchHandle theAuxiliaryObject(timeout); theAuxiliaryObject.cont(); theAuxiliaryObject.next())

#define await_case(cond) if(theAuxiliaryObject.may() && theAuxiliaryObject.check(cond))

#define await_timeout() if(theAuxiliaryObject.timedout())

#define await(cond) await_switch(await_n::FOREVER) await_case(cond)


#define lock_switch(timeout) FOR(await_n::LockSwitchHandle theAuxiliaryObject(timeout, ExtraInfo(__FILE__, __LINE__, __FUNCTION_NAME__)); theAuxiliaryObject.cont(); theAuxiliaryObject.next())

#define lock_case(slot) if(theAuxiliaryObject.prep()) theAuxiliaryObject.collect(slot); else if(theAuxiliaryObject.selected())

#define lock_timeout() if(theAuxiliaryObject.timedout())

#define lock_victim() if(theAuxiliaryObject.victimed())

#define lock_it(slot) lock_switch(await_n::FOREVER) lock_case(slot)


#define rlock(lvalue) await_n::LockSlot::makeSharedLock(&(lvalue))

#define wlock(lvalue) await_n::LockSlot::makeExclusiveLock(&(lvalue))


#endif
//=============================================================================
