/**
 * lockmanager.h
 */
#ifndef __LOCKMANAGER_INCLUDED
#define __LOCKMANAGER_INCLUDED

#include <await/export.hdr>

namespace await_n {

class LockSlotObject;
class AWAIT_API LockSlot {
private:
	LockSlotObject *m_pObject;
	LockSlot(LockSlotObject *);
public:
	LockSlot();
	LockSlot(LockSlot const &);
	~LockSlot();
	LockSlot &operator = (LockSlot const &);
	LockSlot &operator &= (LockSlot const &);
	LockSlot operator && (LockSlot const &) const;
	static LockSlot makeExclusiveLock(void const *pResource);
	static LockSlot makeSharedLock(void const *pResource);
// for internal use
	inline LockSlotObject *getObject() const { return m_pObject; }
};

class LockQueryObject;
class AWAIT_API LockQuery {
private:
	LockQueryObject *m_pObject;
	LockQuery(LockQueryObject *);
public:
	LockQuery();
	LockQuery(LockQuery const &);
	~LockQuery();
	LockQuery &operator = (LockQuery const &);
	LockQuery &operator |= (LockQuery const &);
	friend LockQuery operator || (LockQuery const &, LockQuery const &);
	friend LockQuery tagLock(long tag, LockSlot const &);
// for internal use
	inline LockQueryObject *getObject() const { return m_pObject; }
};

/**
 * Lock
 * temporary object
 */
struct Lock {
	virtual long tag() =0;
	virtual void unlock() =0;
};

/** LockException */
class LockException {};

/** TimeoutException */
class TimeoutException: public LockException {};

/** DeadlockException */
class DeadlockException: public LockException {};

/**
 * Inforamation about sleeping point of the thread
 */
struct ExtraInfo {
	char const *m_szFilename;
	long m_nLine;
	char const *m_szFunction;
	inline ExtraInfo(char const *szFilename, long nLine, char const *szFunction):
		m_szFilename(szFilename), m_nLine(nLine), m_szFunction(szFunction)
	{}
	inline ExtraInfo(ExtraInfo const &o):
		m_szFilename(o.m_szFilename), m_nLine(o.m_nLine), m_szFunction(o.m_szFunction)
	{}
};

/**
 * LockManager
 * singleton
 */
struct LockManager {
	virtual Lock *lock(LockQuery const &, long timeout, ExtraInfo const &) throw(LockException) =0;
	virtual void dumpLockTables() =0;
};

AWAIT_API LockManager *getLockManager();

} // namespace await_n

#endif
//=============================================================================
