/**
 * monitor.h
 */
#ifndef __MONITOR_INCLUDED
#define __MONITOR_INCLUDED

#include <await/export.hdr>

namespace await_n {

#if defined(_MSC_VER)
typedef __int64 AbsoluteMoment;
#else
typedef long long AbsoluteMoment;
#endif

/**
 * Monitor
 */
struct Monitor {
    /** enters critical section */
    virtual void enter() =0;
    /** leaves critical section */
    virtual void leave() =0;
    /** 
     * sleeping, should be invoked inside critical section
     * returns true if timed out.
     */
    virtual bool sleep(AbsoluteMoment) =0;
    /** awakes all sleeping thread, should be invoked inside critical section */
    virtual void awake() =0;
};

//-----------------------------------------------------------------------------

/** returns raw monitor (common for whole application) */
AWAIT_API Monitor *getGlobalMonitor();

/** converts relative time (in milliseconds) to absolute right now */
AWAIT_API AbsoluteMoment cvtRelativeToAbsolute(long);

} // namespace await_n

#endif
//=============================================================================
