/**
 * thread.h
 */
#ifndef __THREAD_INCLUDED
#define __THREAD_INCLUDED

#include <thread/export.hdr>
#include <memory/refcounter.h>

namespace thread {

/**
 * Runnable -- interface that should be implemented by the client of Thread-component.
 */
struct Runnable: virtual public memory::RefCounter {
	/** this method shall be run in the context of another thread */
	virtual void run() =0;
};

/**
 * Thread -- represents thread
 */
struct Thread: virtual public memory::RefCounter {
	/** checks whether thread is finished, should be always invoked inside critical section */
	virtual int _isFinished() =0;
	/** checks whether thread is running, should be always invoked inside critical section */
	inline int _isRunning() { return !_isFinished(); }
};

/** creates new thread */
THREAD_API Thread *createThread(Runnable *);

/** for testing purpose */
THREAD_API void spendSomeTime(long time);

} // namespace thread

#endif
//=============================================================================
