/**
 * lockquery.cpp
 */
#include "i_lock.h"

namespace await_n {

using namespace std;

//-----------------------------------------------------------------------------

LockSlotObject::LockSlotObject(void const *pResource, int nLockType): m_mContent() {
	m_mContent[pResource] = nLockType;
}

LockSlotObject::LockSlotObject(): m_mContent() {
	// do nothing more
}

void LockSlotObject::add(LockSlotObject const *p) {
	if(p) {
		for(UnitedLockTable::const_iterator i = p->m_mContent.begin(), n = p->m_mContent.end(); i != n; ++i) {
			void const *pR = i->first;
			UnitedLockTable::iterator j = m_mContent.find(pR);
			if(j == m_mContent.end()) // not found
				m_mContent[pR] = i->second;
			else
				if(j->second < i->second)
					j->second = i->second;
		}
	}
}

//-----------------------------------------------------------------------------

LockSlot::LockSlot(LockSlotObject *pObject): m_pObject(pObject) {
	// do nothing more
}

LockSlot::LockSlot(): m_pObject(0) {
	m_pObject = new LockSlotObject();
}

LockSlot::LockSlot(LockSlot const &o): m_pObject(0) {
	m_pObject = new LockSlotObject();
	m_pObject->add(o.m_pObject);
}

LockSlot::~LockSlot() {
	delete m_pObject;
}

LockSlot &LockSlot::operator = (LockSlot const &o) {
	if(this != &o) {
		delete m_pObject;
		m_pObject = new LockSlotObject();
		m_pObject->add(o.m_pObject);
	}
	return *this;
}

LockSlot &LockSlot::operator &= (LockSlot const &o) {
	if(m_pObject)
		m_pObject->add(o.m_pObject);
	return *this;
}

LockSlot LockSlot::operator && (LockSlot const &two) const {
	LockSlotObject *pObject = new LockSlotObject();
	pObject->add(m_pObject);
	pObject->add(two.m_pObject);
	return LockSlot(pObject);
}

LockSlot LockSlot::makeExclusiveLock(void const *pResource) {
	return LockSlot(new LockSlotObject(pResource, EXCLUSIVE_LOCK));
}

LockSlot LockSlot::makeSharedLock(void const *pResource) {
	return LockSlot(new LockSlotObject(pResource, SHARED_LOCK));
}

//-----------------------------------------------------------------------------

LockQueryObject::LockQueryObject(long nTag, LockSlotObject *pSlot): m_mContent() {
	m_mContent[nTag] = pSlot->getUnitedLockTable();
}

LockQueryObject::LockQueryObject(): m_mContent() {
	// do nothing more
}

void LockQueryObject::add(LockQueryObject const *p) {
	if(p) {
		for(ChoiceLockTable::const_iterator i = p->m_mContent.begin(), n = p->m_mContent.end(); i != n; ++i) {
			for(UnitedLockTable::const_iterator j = i->second.begin(), m = i->second.end(); j != m; ++j) {
				UnitedLockTable::iterator k = m_mContent[i->first].find(j->first);
				if(k == m_mContent[i->first].end())
					m_mContent[i->first][j->first] = j->second;
				else
					if(k->second < j->second)
						k->second = j->second;
			}
		}
	}
}

//-----------------------------------------------------------------------------

LockQuery::LockQuery(LockQueryObject *pObject): m_pObject(pObject) {
	// do nothing more
}

LockQuery::LockQuery(): m_pObject(0) {
	m_pObject = new LockQueryObject();
}

LockQuery::LockQuery(LockQuery const &o): m_pObject(0) {
	m_pObject = new LockQueryObject();
	m_pObject->add(o.m_pObject);
}

LockQuery::~LockQuery() {
	delete m_pObject;
}

LockQuery &LockQuery::operator = (LockQuery const &o) {
	if(this != &o) {
		delete m_pObject;
		m_pObject = new LockQueryObject();
		m_pObject->add(o.m_pObject);
	}
	return *this;
}

LockQuery &LockQuery::operator |= (LockQuery const &o) {
	if(m_pObject)
		m_pObject->add(o.m_pObject);
	return *this;
}

LockQuery operator || (LockQuery const &one, LockQuery const &two) {
	LockQueryObject *pObject = new LockQueryObject();
	pObject->add(one.m_pObject);
	pObject->add(two.m_pObject);
	return LockQuery(pObject);
}

LockQuery tagLock(long tag, LockSlot const &o) {
	return LockQuery(new LockQueryObject(tag, o.getObject()));
}

} // namespace await_n

//=============================================================================
