/**
 * win32/monitor.cpp
 */
#include "i_monitor.h"

#if _WIN32_WINNT < 0x0400
#define SignalObjectAndWait(m_hMutex, m_hEvent, dwTimeout, bUnused) (ReleaseMutex(m_hMutex), WaitForSingleObject(m_hEvent, dwTimeout))
#endif

namespace await_n {

enum { MAX_ABSOLUTE = 0x7fffffff };

MonitorImpl::MonitorImpl(): m_nCount(0) {
	m_hMutex = CreateMutex(NULL, FALSE, NULL);
	m_hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
}

MonitorImpl::~MonitorImpl() {
	CloseHandle(m_hEvent);
	CloseHandle(m_hMutex);
}

void MonitorImpl::enter() {
	WaitForSingleObject(m_hMutex, INFINITE);
	if(m_nCount>0) // unlock if it had been already locked
		ReleaseMutex(m_hMutex);
	m_nCount++;
}

void MonitorImpl::leave() {
	m_nCount--;
	if(m_nCount == 0)
		ReleaseMutex(m_hMutex);
}

bool MonitorImpl::sleep(AbsoluteMoment moment) {
	moment -= GetTickCount();
	DWORD dwTimeout = (moment >= MAX_ABSOLUTE)? INFINITE: (DWORD) moment;
//	ReleaseMutex(m_hMutex);
//	DWORD res = WaitForSingleObject(m_hEvent, dwTimeout);
	long nCount = m_nCount; // save lock count in thread specific memory
	m_nCount = 0;
	DWORD res = SignalObjectAndWait(m_hMutex, m_hEvent, dwTimeout, FALSE);
	WaitForSingleObject(m_hMutex, INFINITE);
	m_nCount = nCount; // restore lock count for current thread
	return (res == WAIT_TIMEOUT);
}

void MonitorImpl::awake() {
	PulseEvent(m_hEvent);
}

//----------------------------------------------------------------------------

Monitor *getGlobalMonitor() {
	static MonitorImpl g_oRawMonitor;
	return &g_oRawMonitor;
}

AbsoluteMoment cvtRelativeToAbsolute(long s) {
	AbsoluteMoment result = s;
	if(result<0)
		result = MAX_ABSOLUTE;
	else
		result += GetTickCount();
	return result;
}

threadid_t getSelfThreadId() {
	return GetCurrentThreadId();
}

} // namespace await_n

//=============================================================================
