/**
 * posix/thread.cpp
 */
#include <pthread.h>
#include <unistd.h>
#include <thread/thread.h>
#include <await/await.h>

namespace thread {

//-----------------------------------------------------------------------------

class ThreadImpl: public Thread {
private:
	Runnable *m_pRunnable;

	static void *ThreadRoutine(void *);
public:
	ThreadImpl(Runnable *);
	virtual ~ThreadImpl();
	// from Thread
	virtual int _isFinished();
	// own
	ThreadImpl *init();
};

//-----------------------------------------------------------------------------

ThreadImpl::ThreadImpl(Runnable *pRunnable): m_pRunnable(pRunnable) {
	// do nothing more
}

ThreadImpl::~ThreadImpl() {
	// do nothing, m_pRunnable shall be removed in ThreadProc or init
}

int ThreadImpl::_isFinished() {
	return (m_pRunnable == 0);
}

void *ThreadImpl::ThreadRoutine(void *pParam) {
	ThreadImpl *pThis = (ThreadImpl *) pParam;
	pThis->m_pRunnable->run();
	pThis->m_pRunnable->decRefCount();
	await(1)
		pThis->m_pRunnable = 0;
	pThis->decRefCount(); // remove reference passed from init()
	return 0;
}

ThreadImpl *ThreadImpl::init() {
	ThreadImpl *result = this;
	pthread_t thr;
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	incRefCount(); // this reference for ThreadRoutine
	if(pthread_create(&thr, &attr, ThreadRoutine, this) == -1) {
		m_pRunnable->decRefCount(); // remove reference to m_pRunnable, because it will not be used and will not be removed in destructor
		// no need to assign m_pRunnable to 0
		decRefCount(); // ThreadRoutine hasn't run, therefore remove previously passed reference
		decRefCount(); // remove itself if thread hasn't been created
		result = 0;
	}
	pthread_attr_destroy(&attr);
	return result;
}

//-----------------------------------------------------------------------------

THREAD_API Thread *createThread(Runnable *pRunnable) {
	Thread *result = 0;
	if(pRunnable) {
		result = (new ThreadImpl(pRunnable))->init();
	}
	return result;
}

THREAD_API void spendSomeTime(long millisec) {
	usleep(millisec*1000);
}

} // namespace thread

//=============================================================================
