// main.cpp
#include <stdio.h>
#include <thread/thread.h>
#include <await/await.h>

using namespace thread;
using namespace memory;
using namespace await_n;

// here are synchronizing flags to share amoung threads
// this flag is to set by consumer
bool volatile g_bCanWrite = false;
// this flag is to set by producer
bool volatile g_bCanRead = false;

struct Producer: virtual public RefCounter, public Runnable {
	virtual void run() {
		for(int i=0; i<3; ++i) { // produce something 3 times
			await(g_bCanWrite) // wait for ability to produce something
				g_bCanWrite = false; // and clear this flag
			fprintf(stderr, "producing...");
			spendSomeTime(1000);
			fprintf(stderr, "ok\n");
			await(true)
				g_bCanRead = true; // notify consumer that data are ready
		}
	}
};

struct Consumer: virtual public RefCounter, public Runnable {
	virtual void run() {
		while(true) {
			int bCanQuit = false;
			await(true)
				g_bCanWrite = true; // notify producer about ability to produce
			await_switch(2000) { // wait at most 2 seconds
				await_case(g_bCanRead) // wait for ability to consume produced data
					g_bCanRead = false;
				await_timeout() { // quit when timed out 
					bCanQuit = true;
					fprintf(stderr, "timed out\n");
				}
			}
			if(bCanQuit)
				break;
			fprintf(stderr, "consuming...");
			spendSomeTime(1000);
			fprintf(stderr, "ok\n");
		}
	}
};

int main() {
	// create threads
	Thread *pConsumer = createThread(new Consumer());
	Thread *pProducer = createThread(new Producer());
	// wait until threads terminated
	await(pConsumer->_isFinished() && pProducer->_isFinished()) {}
	// remove references to threads
	pConsumer->decRefCount();
	pProducer->decRefCount();

	return 0;
}

//=============================================================================
