// main.cpp
#include <stdio.h>
#include <thread/thread.h>
#include <await/await.h>

using namespace thread;
using namespace memory;
using namespace await_n;

enum { NUMBER_OF_THREADS = 2 };
enum { FINAL_POINT = 4 };

int g_nPoint = -1;
void point(int i) {
	await(g_nPoint >= i) {}
}

// here is shared resources
int res0;
double res1;
char res2;

struct One: virtual public RefCounter, public Runnable {
	virtual void run() {
		point(0);
		fprintf(stderr, "one: shared lock for res0 and exclusive lock for res2, acquiring\n");
		lock_it(rlock(res0) && wlock(res2)) {
			fprintf(stderr, "one: shared lock for res0 and exclusive lock for res2, acquired\n");
			point(1);
			fprintf(stderr, "one: exclusive lock for res1, acquiring\n");
			lock_it(wlock(res1)) {
				fprintf(stderr, "one: exclusive lock for res1, acquired\n");
				point(3);
				fprintf(stderr, "one: exclusive lock for res1, released\n");
			}
			fprintf(stderr, "one: shared lock for res0 and exclusive lock for res2, released\n");
		}
	}
};

struct Two: virtual public RefCounter, public Runnable {
	virtual void run() {
		point(0);
		fprintf(stderr, "two: shared lock for res1, acquiring\n");
		lock_it(rlock(res1)) {
			fprintf(stderr, "two: shared lock for res1, acquired\n");
			point(2);
			fprintf(stderr, "two: exclusive lock for res0 or shared lock for res2, acquiring\n");
			lock_switch(FOREVER) {
				lock_case(wlock(res0)) {
					fprintf(stderr, "two: exclusive lock for res0, acquired\n");
					point(3);
					fprintf(stderr, "two: exclusive lock for res0, released\n");
				}
				lock_case(rlock(res2)) {
					fprintf(stderr, "two: shared lock for res2, acquired\n");
					point(3);
					fprintf(stderr, "two: shared lock for res2, released\n");
				}
				lock_victim() {
					fprintf(stderr, "deadlock\n");
					point(3);
				}
			}
			fprintf(stderr, "two: shared lock for res1, released\n");
		}
	}
};

bool _isAllFinished(Thread **ppThread, int n) {
	for(int i=0; i<n; ++i)
	if(ppThread[i]->_isRunning())
		return false;
	return true;
}

int main() {
	int i;
	Thread *ppThread[NUMBER_OF_THREADS];
	// create threads
	ppThread[0] = createThread(new One());
	ppThread[1] = createThread(new Two());
	// pulse
	for(i=0; i<=FINAL_POINT; ++i) {
		spendSomeTime(1000);
		getLockManager()->dumpLockTables();
		fprintf(stderr, "--> %d\n", i);
		await(true)
			g_nPoint = i;
	}
	// wait until threads terminated
	await(_isAllFinished(ppThread, NUMBER_OF_THREADS)) {}
	// remove references to threads
	for(i=0; i<NUMBER_OF_THREADS; ++i)
		ppThread[i]->decRefCount();

	return 0;
}

//=============================================================================
