// main.cpp
#include <stdio.h>
#include <thread/thread.h>
#include <await/await.h>

using namespace thread;
using namespace memory;
using namespace await_n;

enum { NUMBER_OF_THREADS = 3 };
enum { FINAL_POINT = 5 };

int g_nPoint = -1;
void point(int i) {
	await(g_nPoint >= i) {}
}

// here is shared resources
int res0;
double res1;

struct One: virtual public RefCounter, public Runnable {
	virtual void run() {
		point(0);
		fprintf(stderr, "one: shared lock for res0, acquiring\n");
		lock_it(rlock(res0)) {
			fprintf(stderr, "one: shared lock for res0, acquired\n");
			point(4);
			fprintf(stderr, "one: shared lock for res0, released\n");
		}
	}
};

struct Two: virtual public RefCounter, public Runnable {
	virtual void run() {
		point(0);
		fprintf(stderr, "two: shared lock for res1, acquiring\n");
		lock_it(rlock(res1)) {
			fprintf(stderr, "two: shared lock for res1, acquired\n");
			point(4);
			fprintf(stderr, "two: shared lock for res1, released\n");
		}
	}
};

struct Three: virtual public RefCounter, public Runnable {
	virtual void run() {
		point(1);
		fprintf(stderr, "three: exclusive lock for res0 or exclusive lock for res1 and shared for res0, acquiring\n");
		lock_switch(3000){
			lock_case(wlock(res0)) {
				fprintf(stderr, "three: exclusive lock for res0, acquired\n");
				point(4);
				fprintf(stderr, "three: exclusive lock for res0, released\n");
			}
			lock_case(wlock(res1) && rlock(res0)) {
				fprintf(stderr, "three: exclusive lock for res1 and shared for res0, acquired\n");
				point(4);
				fprintf(stderr, "three: exclusive lock for res1 and shared for res0, released\n");
			}
			lock_timeout() {
				fprintf(stderr, "three: time out\n");
			}
		}
	}
};

bool _isAllFinished(Thread **ppThread, int n) {
	for(int i=0; i<n; ++i)
	if(ppThread[i]->_isRunning())
		return false;
	return true;
}

int main() {
	int i;
	Thread *ppThread[NUMBER_OF_THREADS];
	// create threads
	ppThread[0] = createThread(new One());
	ppThread[1] = createThread(new Two());
	ppThread[2] = createThread(new Three());
	// pulse
	for(i=0; i<=FINAL_POINT; ++i) {
		spendSomeTime(1000);
		getLockManager()->dumpLockTables();
		fprintf(stderr, "--> %d\n", i);
		await(true)
			g_nPoint = i;
	}
	// wait until threads terminated
	await(_isAllFinished(ppThread, NUMBER_OF_THREADS)) {}
	// remove references to threads
	for(i=0; i<NUMBER_OF_THREADS; ++i)
		ppThread[i]->decRefCount();

	return 0;
}

//=============================================================================
